import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:slide_container/slide_container.dart';
import 'package:slide_container/slide_container_controller.dart';

/// Use this instead of a [SingleChildScrollView] when the ScrollView is inside a [SlideContainer].
/// This custom [SingleChildScrollView] detects over-scroll and calls a forceSlide on the parent [SlideContainer]
/// to allow the [SlideContainer] to slide.
///
/// A normal [SingleChildScrollView] would consume all the slide gesture events, preventing the parent [SlideContainer]
/// to ever detect them.
class SlideContainerScrollView extends StatefulWidget {
  /// Will be used to trigger a force slide when the user over-scroll the ScrollView.
  final SlideContainerController slideContainerController;

  /// Min distance the users have to over-scroll toward the outside of the ScrollView to trigger
  /// the SlideContainer forced slide.
  ///
  /// In pixels.
  final double minTotalOverScrollToTriggerSlideContainer;

  /// Copied from SingleChildScrollView.
  final Axis scrollDirection;
  final bool reverse;
  final EdgeInsetsGeometry padding;
  final ScrollController controller;
  final Widget child;
  final DragStartBehavior dragStartBehavior;

  const SlideContainerScrollView({
    @required this.slideContainerController,
    this.minTotalOverScrollToTriggerSlideContainer = 150,
    Key key,
    this.scrollDirection = Axis.vertical,
    this.reverse = false,
    this.padding,
    this.controller,
    this.child,
    this.dragStartBehavior = DragStartBehavior.start,
  })  : assert(scrollDirection != null),
        assert(minTotalOverScrollToTriggerSlideContainer != null),
        assert(slideContainerController != null),
        assert(dragStartBehavior != null),
        super(key: key);

  @override
  _SlideContainerScrollViewState createState() =>
      _SlideContainerScrollViewState();
}

class _SlideContainerScrollViewState extends State<SlideContainerScrollView> {
  ScrollController scrollController;
  bool startedScrollingAtMinScrollPosition = false;
  bool startedScrollingAtMaxScrollPosition = false;
  bool isOnlyOverScroll = false;
  double totalOverScroll = 0;
  bool isOverScrollDirectionPositive = true;

  bool get startedScrollingAtEdge =>
      startedScrollingAtMinScrollPosition ||
      startedScrollingAtMaxScrollPosition;

  SlideContainerController get slideContainerController =>
      widget.slideContainerController;

  @override
  void initState() {
    scrollController = widget.controller ?? ScrollController();
    super.initState();
  }

  @override
  void dispose() {
    if (widget.scrollDirection == null) scrollController.dispose();
    super.dispose();
  }

  bool onScrollNotification(ScrollNotification notification) {
    if (notification is ScrollStartNotification) {
      totalOverScroll = 0;
      startedScrollingAtMinScrollPosition = false;
      startedScrollingAtMaxScrollPosition = false;
      if (notification.metrics.atEdge) {
        isOnlyOverScroll = true;
        if (notification.metrics.pixels == 0.0) {
          startedScrollingAtMinScrollPosition = true;
        } else {
          startedScrollingAtMaxScrollPosition = true;
        }
      }
    }

    if (notification is ScrollUpdateNotification) {
      isOnlyOverScroll = false;
    }

    if (notification is OverscrollNotification) {
      if (isOnlyOverScroll && startedScrollingAtEdge) {
        totalOverScroll += notification.overscroll.abs();
        isOverScrollDirectionPositive =
            notification.overscroll > 0 && !widget.reverse;
        if (totalOverScroll >
            widget.minTotalOverScrollToTriggerSlideContainer) {
          /// This jumpTo is to cancel the current scroll event and allow the SlideContainer to slide.
          scrollController.jumpTo(scrollController.offset);
          if (notification.metrics.axis == Axis.vertical) {
            if (isOverScrollDirectionPositive) {
              slideContainerController
                  .forceSlide(SlideContainerDirection.bottomToTop);
            } else {
              slideContainerController
                  .forceSlide(SlideContainerDirection.topToBottom);
            }
          } else {
            if (isOverScrollDirectionPositive) {
              slideContainerController
                  .forceSlide(SlideContainerDirection.rightToLeft);
            } else {
              slideContainerController
                  .forceSlide(SlideContainerDirection.leftToRight);
            }
          }
        }
      }
    }

    return false;
  }

  @override
  Widget build(BuildContext context) =>
      NotificationListener<ScrollNotification>(
        onNotification: onScrollNotification,
        child: SingleChildScrollView(
          /// Physics needs to be ClampingScrollPhysics as the default iOS bouncing over-scroll behaviour doesn't work for our purpose.
          physics: ClampingScrollPhysics(),
          controller: scrollController,
          primary: false,
          scrollDirection: widget.scrollDirection,
          reverse: widget.reverse,
          key: widget.key,
          padding: widget.padding,
          child: widget.child,
          dragStartBehavior: widget.dragStartBehavior,
        ),
      );
}
